# Import tkinter and filedialog
import random
import sys
import threading
import tkinter as tk

# Import python standard packages
from time import sleep
import os
from tkinter import filedialog as tkfile_dialog

# Importing from another file works differently than if running this file directly for some reason.
# Import all required modules from nebulatk
try:
    from . import (
        bounds_manager,
        fonts_manager,
        image_manager,
        initialize,
        standard_methods,
    )
except ImportError:
    import bounds_manager
    import fonts_manager
    import image_manager
    import initialize
    import standard_methods


# Initialize base methods for all widgets.
# This is largely so we don't ever need to initialize methods that will never be used (e.g. hovered on a frame)
class _widget:

    def hovered(self):
        pass

    def hover_end(self):
        pass

    def clicked(self):
        pass

    def dragging(self, x, y):
        pass

    def release(self):
        pass

    def destroy(self):
        standard_methods.delete(self)

    def typed(self, character):
        pass

    def change_active(self):
        pass

    def _update_children(self, children=None, command="update"):
        if children is None:
            children = self.children
        if children != []:
            for child in children:
                getattr(child, command)()
                self._update_children(child.children)

    def _hide(self):
        """Hide the widget"""

        # Flop off hides all items that are part of this widget
        standard_methods.flop_off(self)

        # Remove this widget from the bounds
        bounds_manager.remove_bounds(self, self.bounds_type)

        # Always return self on methods the user might call. this allows for chaining like button = Button().place().hide()
        return self

    def hide(self):
        self.visible = False
        self._hide()
        self._update_children(command="_hide")
        return self

    def _show(self):
        """Shows the widget"""
        # Show makes all items that are part of this widget visible
        standard_methods.flop_on(self)

        # Add this widget to bounds
        bounds_manager.update_bounds(self, self.x, self.y, self.bounds_type)
        return self

    def show(self):
        self.visible = True
        self._show()
        self._update_children(command="_show")
        return self

    # Default place behavior
    def place(self, x=0, y=0):
        """Place the widget at the specified position.

        Args:
            x (int, optional): x position. Defaults to 0.
            y (int, optional): y position. Defaults to 0.
        """
        self.abs_x = x
        self.abs_y = y
        x = int(x)
        y = int(y)
        if self.bg_object is None and self.image_object is None:
            standard_methods.place_bulk(self, x, y)
            if self.bg_object is not None and self.bounds_type == "box":
                self.object = self.bg_object
            elif self.image_object is not None:
                self.object = self.image_object
            else:
                self.object = self.text_object
        else:
            standard_methods.update_positions(self, x, y)
        bounds_manager.update_bounds(self, x, y, mode=self.bounds_type)

        self.x = x
        self.y = y

        self._update_children()

        if hasattr(self, "original_x"):
            self.original_x = x
            self.original_y = y

        return self

    def update(self):
        standard_methods.delete(self)
        self.place(self.x, self.y)

    # Default configure behavior
    def configure(
        self, x=None, y=None, width=None, height=None, text=None, fill="Default"
    ):
        """Configure the widget with the specified parameters

        Args:
            x (int, optional): x position. Defaults to None.
            y (int, optional): y position. Defaults to None.
            text (str, optional): Text. Defaults to None.
            fill (color, optional): Fill color. Defaults to "Default".
        """
        if x is not None:
            # Update x position of all items in this widget
            standard_methods.update_positions(self, x, self.y)

        if y is not None:
            # Update y position of all items in this widget
            standard_methods.update_positions(self, self.x, y)

        if text is not None:
            # Update text in this widget
            if self.text_object is not None:
                self.master.configure(self.text_object, text=text)

            if self.active_text_object is not None:
                self.master.configure(self.active_text_object, text=text)

            self.text = text

        if fill != "Default":
            self.fill = fill
            self.update()

        if width is not None:
            self.width = width
            self.update()

        if height is not None:
            self.height = height
            self.update()

        # Update the bounds for this widget
        bounds_manager.update_bounds(self, x, y, mode=self.bounds_type)

        if x is not None:
            self.x = x
        if y is not None:
            self.y = y

        if hasattr(self, "original_x"):
            self.original_x = x
            self.original_y = y
        return self


# Our implementation of tkinter's file_dialog.
# This mostly just exists so that the user doesn't have to import nebulatk and tkinter
def FileDialog(window, initialdir=None, mode="r", filetypes=(("All files", "*"))):
    """Identical to tkinter.FileDialog

    Args:
        window (nebulatk.Window): Root window
        initialdir (str, optional): Initial directory to open to. Defaults to None.
        mode (str, optional): File open mode. Defaults to "r".
        filetypes (list, optional): List of filetypes. Format like:
            [
                ("Name","*.ext"),
                ("Name", ("*.ext", "*.ext2"))
            ]
            Defaults to [("All files", "*")].

    Returns:
        file: Open file
    """
    window.leave_window(None)
    file = tkfile_dialog.askopenfile(
        initialdir=initialdir,
        mode=mode,
        filetypes=filetypes,
    )
    window.leave_window(None)
    return file


class Button(_widget):

    def __init__(
        self,
        root=None,
        width=0,
        height=0,
        text="",
        font=None,
        justify="center",
        text_color="default",
        active_text_color=None,
        fill="default",
        active_fill="default",
        border="default",
        border_width=0,
        image=None,
        active_image=None,
        hover_image=None,
        hover_image_active=None,
        bounds_type="default",
        command=None,
        command_off=None,
        state=False,
        mode="standard",
    ):
        """NebulaTk Button widget.

        Args:
            root (nebulatk.Window, optional): Root Window. Defaults to None.

            width (int, optional): width. Defaults to 0.
            height (int, optional): height. Defaults to 0.

            text (str, optional): text. Defaults to "".
            font (_type_, optional): Font. Font name, or font tuple. Defaults to None.
            justify (str, optional): Justification for text. Defaults to None.

            text_color (str, color): text color. Defaults to "default".
            active_text_color (color, optional): color. Defaults to None.

            fill (color, optional): Fill color. Defaults to "white".
            active_fill (color, optional): Active fill color

            border (color, optional): Border color. Defaults to "black".
            border_width (int, optional): Border width. Defaults to 0.

            image (str, optional): Image path. Defaults to None.
            active_image (str, optional): Image path. Defaults to None.
            hover_image (str, optional): Image path. Defaults to None.
            hover_image_active (str, optional): Image path. Defaults to None.

            bounds_type (str, optional): Defaults to "box" if image is not provided, or "nonstandard" otherwise.

            command (function, optional): Defaults to None.
            command_off (function, optional): Defaults to None.

            state (bool, optional): _description_. Defaults to False.
            mode (str, optional): _description_. Defaults to "standard".
        """

        # mode = "toggle" is equivalent to a tkinter checkbutton

        # Load most initial variables
        initialize.load_initial(
            self,
            root,
            width,
            height,
            border_width,
            justify,
            state,
            mode,
            command,
            command_off,
        )

        # Load our bounds type
        initialize.load_bounds_type(self, bounds_type, image)

        # Load all of our images
        initialize.load_bulk_images(
            self,
            image=image,
            active_image=active_image,
            hover_image=hover_image,
            hover_image_active=hover_image_active,
        )

        # Load and initialize our text and font
        initialize.load_text(self, text, font)

        # Load all of our colors
        initialize.load_all_colors(
            self, fill, active_fill, border, text_color, active_text_color
        )

        # Convert all of our colors to hexadecimal
        initialize.convert_all_colors(self)

    # Utilize our standard methods to manage hovering
    def hovered(self):
        if self.mode == "toggle":
            standard_methods.hovered_toggle(self)
        elif self.mode == "standard":
            standard_methods.hovered_standard(self)
        self.hovering = True

    def hover_end(self):
        standard_methods.hover_end(self)
        self.hovering = False

    # Utilize our standard methods to manage clicking
    def clicked(self):
        if self.mode == "toggle":
            standard_methods.clicked_toggle(self)

        elif self.mode == "standard":
            standard_methods.clicked_standard(self)

    # No standard method for releasing, most other widgets don't use this method
    def release(self):
        # We only need to do something if this == a standard button, and not a toggle button
        if self.mode == "standard":
            # Toggle state
            self.state = not self.state

            if self.visible:
                if self.hovering:
                    # image_flop hides all other images and shows the selected image
                    standard_methods.image_flop(self, "hover_object")
                else:
                    standard_methods.image_flop(self, "image_object")

                if self.command_off is not None:
                    self.command_off()

    # Utilize our standard methods to manage toggling
    def toggle(self):
        """Toggle appearance of button"""
        if self.mode == "toggle":
            standard_methods.toggle_object_toggle(self)

        elif self.mode == "standard":
            standard_methods.toggle_object_standard(self)


class Slider(_widget):

    def __init__(
        self,
        root=None,
        width=0,
        height=0,
        slider_height=10,
        minimum=0,
        maximum=100,
        text="",
        font=None,
        justify="center",
        text_color="default",
        active_text_color=None,
        fill="default",
        active_fill="default",
        border="default",
        border_width=0,
        slider_border_width=0,
        slider_fill="default",
        slider_border="default",
        image=None,
        active_image=None,
        hover_image=None,
        hover_image_active=None,
        bounds_type="default",
        command=None,
        command_off=None,
        state=False,
    ):
        """Slider widget

        Args:
            root (nebulatk.Window, optional): Root Window. Defaults to None.

            width (int, optional): width. Defaults to 0.
            height (int, optional): height. Defaults to 0.

            slider_height (int, optional): Slider height. Defaults to 10.

            minimum (int, optional): Minimum slider value. Defaults to 0.
            maximum (int, optional): Maximum slider value. Defaults to 100.

            text (str, optional): Defaults to "".
            font (_type_, optional): Font. Font name, or font tuple. Defaults to None.
            justify (str, optional): Justification for text. Defaults to None.

            text_color (str, color): color. Defaults to "default".
            active_text_color (color, optional): color. Defaults to None.

            fill (color, optional): Fill color. Defaults to "white".
            active_fill (color, optional): Active fill color

            border (color, optional): Border color. Defaults to "black".

            border_width (int, optional): Border width. Defaults to 0.
            slider_border_width (int, optional): Border width. Defaults to 0.

            slider_fill (color, optional): Slider fill color. Defaults to "default".
            slider_border (color, optional): Border color. Defaults to "black".

            image (str, optional): Image path. Defaults to None.
            active_image (str, optional): Image path. Defaults to None.
            hover_image (str, optional): Image path. Defaults to None.
            hover_image_active (str, optional): Image path. Defaults to None.

            bounds_type (str, optional): Defaults to "box" if image is not provided, or "nonstandard" otherwise.

            command (function, optional): Defaults to None.
            command_off (function, optional): Defaults to None.

            state (bool, optional): _description_. Defaults to False.
        """

        # Load most initial variables
        initialize.load_initial(
            self,
            root,
            width,
            height,
            border_width,
            justify,
            state,
            None,
            command,
            command_off,
            slider_height,
            maximum,
            minimum,
            slider_border_width,
        )

        # Load bounds type
        initialize.load_bounds_type(self, bounds_type, image)

        # Load all of our images
        initialize.load_bulk_images(
            self,
            image=image,
            active_image=active_image,
            hover_image=hover_image,
            hover_image_active=hover_image_active,
        )

        # Load and initialize our text and font
        initialize.load_text(self, text, font)

        # Load all of our colors
        initialize.load_all_colors(
            self,
            fill,
            active_fill,
            border,
            text_color,
            active_text_color,
            slider_fill,
            slider_border,
        )

        # Convert all of our colors to hexadecimal
        initialize.convert_all_colors(self)

    # Utilize our standard methods for hovering
    def hovered(self):
        standard_methods.hovered_standard(self)
        self.hovering = True

    def hover_end(self):
        standard_methods.hover_end(self)
        self.hovering = False

    # Utilize standard method for toggling
    def toggle(self):
        standard_methods.toggle_object_standard(self)

    # Implement dragging along x ax== for slider
    def dragging(self, x, y):
        # Change our x position to be on the edge of the slider
        # this has the effect of the mouse moving the slider from the center of the slider
        x = x - self.width / 2

        # Clamp our x to the minimum and maximum values, compensating for the position offset,
        x = standard_methods.clamp(
            x, self.minimum + self.original_x, self.maximum + self.original_x
        )

        # Update the positions of all the objects in the slider, avoiding updating the actual background and widget positions
        standard_methods.update_positions(self, x, self.y, avoid_slider=True)

        # Update the bounds of the slider
        bounds_manager.update_bounds(self, x, self.y, mode=self.bounds_type)

        # Execute whatever command == set for dragging
        if self.command:
            self.command()

        # Update our x position
        self.x = x


class Label(_widget):

    def __init__(
        self,
        root=None,
        width=0,
        height=0,
        text="",
        font=None,
        justify="center",
        text_color="default",
        fill="default",
        border="default",
        border_width=0,
        image=None,
        bounds_type="default",
        angle=0,
    ):
        """_summary_

        Args:
            root (nebulatk.Window, optional): Root Window. Defaults to None.
            width (int, optional): width. Defaults to 0.
            height (int, optional): height. Defaults to 0.
            text (str, optional): Defaults to "".
            font (_type_, optional): Font. Font name, or font tuple. Defaults to None.
            fill (color, optional): Fill color. Defaults to "white".
            border (str, optional): Border color. Defaults to "black".
            border_width (int, optional): Border width. Defaults to 3.
            image (str, optional): Image path. Defaults to None.
            bounds_type (str, optional): _description_. Defaults to "box" if image is not provided, or "nonstandard" otherwise. Defaults to "box" if no image == provided, or "non-standard" if an image == provided.
        """
        # Load most initial variables
        initialize.load_initial(
            self,
            root,
            width,
            height,
            border_width,
            justify,
        )

        # Load our bounds type
        initialize.load_bounds_type(self, bounds_type, image)

        # Load all of our images
        initialize.load_bulk_images(self, image=image)

        # Load and initialize our text and font
        initialize.load_text(self, text, font)

        # Load all of our colors
        initialize.load_all_colors(
            self, fill=fill, border=border, text_color=text_color
        )

        # Convert all of our colors to hexadecimal
        initialize.convert_all_colors(self)

        self.angle = angle


class Entry(_widget):

    def __init__(
        self,
        root=None,
        width=0,
        height=0,
        text="",
        font=None,
        justify="center",
        text_color="default",
        fill="default",
        border="default",
        border_width=0,
        image=None,
        bounds_type="default",
    ):
        """_summary_

        Args:
            root (nebulatk.Window, optional): Root Window. Defaults to None.
            width (int, optional): width. Defaults to 0.
            height (int, optional): height. Defaults to 0.
            text (str, optional): Defaults to "".
            font (_type_, optional): Font. Font name, or font tuple. Defaults to None.
            fill (color, optional): Fill color. Defaults to "white".
            border (str, optional): Border color. Defaults to "black".
            border_width (int, optional): Border width. Defaults to 3.
            image (str, optional): Image path. Defaults to None.
            bounds_type (str, optional): Defaults to "box" if image is not provided, or "nonstandard" otherwise.
        """
        # Load most initial variables
        initialize.load_initial(
            self,
            root,
            width,
            height,
            border_width,
            justify,
        )

        # Load our bounds type
        initialize.load_bounds_type(self, bounds_type, image)

        # Load all of our images
        initialize.load_bulk_images(self, image=image)

        # Load and initialize our text and font
        initialize.load_text(self, text, font)

        # Load all of our colors
        initialize.load_all_colors(
            self, fill=fill, border=border, text_color=text_color
        )

        # Convert all of our colors to hexadecimal
        initialize.convert_all_colors(self)

    def clicked(self):
        self.active = True

    def typed(self, character):
        # We will be using self.entire_text, not self.text, as self.configure will change self.text to be the displayed slice of self.entire_text
        # Backspace character
        if character == "\x08":
            self.entire_text = self.entire_text[:-1]

        # If the character is a visible character
        elif character in fonts_manager.ALPHANUMERIC_PLUS:
            self.entire_text = self.entire_text + character

        # Move end of display to the end of the text
        self.end = len(self.entire_text)

        # Get maximum length of characters we can fit in the widget
        max_length = fonts_manager.get_max_length(
            self.master, self.entire_text, self.font, self.width, self.end
        )

        # Configure display text to be the slice of text that we can fit in the widget
        if self.text_object is not None:
            self.configure(
                text=self.entire_text[self.end - max_length : self.end],
            )
        else:
            self.text = self.entire_text[self.end - max_length : self.end]
            self.update()

    def get(self):
        return self.entire_text


class Frame(_widget):

    def __init__(
        self,
        root=None,
        width=0,
        height=0,
        image=None,
        fill=None,
        border="black",
        border_width=1,
        bounds_type="box",
    ):
        """_summary_

        Args:
            root (nebulatk.Window, optional): Root Window. Defaults to None.
            width (int, optional): width. Defaults to 0.
            height (int, optional): height. Defaults to 0.
            image (str, optional): Image path. Defaults to None.
            fill (color, optional): Fill color. Defaults to None.
            border (str, optional): Border color. Defaults to "black".
            border_width (int, optional): Border width. Defaults to 1.
            bounds_type (str, optional): Defaults to "box" if image is not provided, or "nonstandard" otherwise. Defaults to "box".
        """

        # Load most initial variables
        initialize.load_initial(self, root, width, height, border_width)

        # Load our bounds type
        initialize.load_bounds_type(self, bounds_type)

        # Load all of our images
        initialize.load_bulk_images(self, image=image)

        # Load all of our colors
        initialize.load_all_colors(self, fill=fill, border=border)

        # Convert all of our colors to hexadecimal
        initialize.convert_all_colors(self)


# Internal window class to implement threading
# NOTE: Threading and tcl do not combine well, so we have to do a lot of stuff ourselves and be careful that all modifications to the tcl windows are done in the same thread
class _window_internal(threading.Thread):

    def __init__(
        self,
        width=500,
        height=500,
        title="ntk",
        canvas_width="default",
        canvas_height="default",
        closing_command=None,
        resizable=(True, True),
        override=False,
    ):
        # Initialize the thread
        super().__init__()

        # Initialize the bounds
        # Bounds == structured as:
        # bounds[y] = [
        #   [_object, start_x, end_x],
        #   [_object2, start_x, end_x],
        # ]
        self.bounds = {}

        # Initialize variables
        self.root = None
        self.master = self
        self.canvas = None
        self.down = None
        self.active = None
        self.hovered = None
        self.width = int(width)
        self.height = int(height)
        self.resizable = resizable
        self.override = override

        self.title = title

        # Initialize canvas size
        if canvas_height == "default":
            canvas_height = self.height
        if canvas_width == "default":
            canvas_width = self.width

        self.canvas_width = canvas_width
        self.canvas_height = canvas_height

        # Because of threading and garbage collection, images created will sometimes have to be stored in this variable to ensure they do not get deleted
        self.images = []

        # Initialize rest of variables
        self.running = True
        self.closing_command = closing_command

    # NOTE: EVENT HANDLERS

    # Handle mouse clicked down event
    def click(self, event):
        # Mouse position == a float, so convert to integer
        x = int(event.x)
        y = int(event.y)

        # Temp variable active_new
        # Default to None in case no object was clicked on
        active_new = None

        # Find the object that was clicked
        # Check if there are any objects initialized at that position
        if y in self.bounds:
            # Iterate over all objects at that y coordinate
            for i in self.bounds[y]:
                # Check if the mouse was within that object's boundary
                if x <= i[2] and x >= i[1]:
                    # Object found
                    # Set active_new and break
                    active_new = i[0]
                    break

        # If the new object == actually new, update the current active widget
        # self.active == used for things like detecting what widget to send keypresses to
        if active_new is not self.active:
            self.active = active_new

        # If the new object wasn't already being clicked
        if active_new is not self.down:
            self.down = active_new

            # If the new object == a valid widget
            if active_new is not None:
                # Click on the object
                active_new.clicked()

    # Handle mouse click up events
    def click_up(self, event):
        # If something was being clicked on, release the widget
        if self.down:
            self.down.release()
            self.down = None

    # Handle mouse movement event
    def hover(self, event):
        # Mouse position == a float, so convert to integer
        x = int(event.x)
        y = int(event.y)

        # Temp variable hovered_new
        # Default to None in case no object was hovered over
        hovered_new = None

        # If something == already being clicked on, this == also a dragging event on that widget.
        if self.down is not None:
            self.down.dragging(x, y)

        # Find the object that was hovered over
        # Check if there are any objects initialized at that position
        if y in self.bounds:
            # Iterate over all objects at that y coordinate
            for i in self.bounds[y]:
                # Check if the mouse was within that object's boundary
                if x <= i[2] and x >= i[1]:
                    # Object found
                    # If this object wasn't already being hovered over, trigger new object's hovered event
                    if i[0] is not self.hovered:
                        i[0].hovered()

                    # Set hovered_new and break
                    hovered_new = i[0]
                    break

        # If the object wasn't already being hovered over, check that there was something being hovered over previously and trigger the old object's hover_end event
        if hovered_new is not self.hovered:
            if self.hovered is not None:
                self.hovered.hover_end()

            # Update the currently hovered object
            self.hovered = hovered_new

    # Handle the mouse leaving the window
    def leave_window(self, event):
        # If something was being hovered, trigger the hover_end event
        if self.hovered is not None:
            self.hovered.hover_end()

        # If something was being clicked on, simulate the mouse button being released
        if self.down is not None:
            self.down.release()

        # Ensure that the variables for selected, clicked, and hovered objects are reset
        self.hovered = None
        self.down = None
        self.active = None

    # Handle keystrokes
    def typing(self, event):
        # If a widget == active, we should send keystrokes to it
        if self.active:
            self.active.typed(event.char)

    # NOTE: CREATION HANDLERS
    # These are necessary so that threading works properly with tcl

    # Wrapper for canvas.create_image method
    def create_image(self, x, y, image, state="normal"):
        self.images.append(image)
        return self.canvas.create_image(x, y, image=image, anchor="nw", state=state)

    # Wrapper for canvas.create_rectangle method
    def create_rectangle(
        self, x, y, widt, height=0, fill=0, border_width=0, outline=None, state="normal"
    ):
        # To support transparency with RGBA, we need to check whether the rectangle includes transparency
        if fill is not None and len(fill) > 7:
            bg_image = image_manager.create_image(
                fill, widt, height, outline, border_width, self
            )
            self.images.append(bg_image)
            return self.create_image(x, y, bg_image, state=state)

        # Otherwise we can continue with creating the rectangle
        return self.canvas.create_rectangle(
            x + border_width / 2,
            y + border_width / 2,
            widt - border_width / 2,
            height - border_width / 2,
            fill=fill,
            width=border_width,
            outline=outline,
            state=state,
        )

    # Wrapper for canvas.create_text method
    def create_text(
        self, x, y, text, font, fill="black", anchor="center", state="normal", angle=0
    ):
        return self.canvas.create_text(
            x,
            y,
            text=text,
            font=font,
            fill=fill,
            anchor=anchor,
            state=state,
            angle=angle,
        )

    # Wrapper for canvas.move method
    def move(self, _object, x, y):
        if _object is not None:
            self.canvas.move(_object, x, y)

    # Wrapper for canvas.delete method
    def delete(self, _object):
        self.canvas.delete(_object)

    # Wrapper for canvas.change_state method
    def change_state(self, _object, state):
        self.canvas.itemconfigure(_object, state=state)

    # Wrapper for canvas.configure method
    def configure(self, _object, **kwargs):
        self.canvas.itemconfigure(_object, kwargs)

    # NOTE: Other methods

    # Handle window closing
    def close(self):
        self.running = False
        try:
            self.root.update()
            self.join()
        except Exception as e:
            print(e)
        if self.closing_command is not None:
            self.closing_command()

    def destroy(self):
        self.close()

    # Add in window.place() to simplify tcl's root.geometry method
    def place(self, x=0, y=0):
        self.root.geometry(f"{self.width}x{self.height}+{x}+{y}")
        return self

    # Wrapper for root.bind() method, in future, add global keypress handling
    def bind(self, key, command):
        self.root.bind(key, command)

    # Main method
    def run(self):
        # Create window
        self.root = tk.Tk()
        self.canvas = tk.Canvas(
            self.root,
            width=self.canvas_width,
            height=self.canvas_height,
            borderwidth=0,
            highlightthickness=0,
        )
        self.canvas.pack()

        # Initialize window
        self.root.geometry(f"{self.width}x{self.height}")

        self.root.title(self.title)

        self.root.resizable(self.resizable[0], self.resizable[1])

        self.root.overrideredirect(self.override)

        def close():
            self.running = False
            try:
                self.root.quit()
            except Exception as e:
                print(f"exception{e}")
            if self.closing_command is not None:
                self.closing_command()

        # Bind events to our new event handlers
        self.canvas.bind("<Button-1>", self.click)
        self.canvas.bind("<ButtonRelease-1>", self.click_up)
        self.canvas.bind_all("<Key>", self.typing)
        self.root.protocol("WM_DELETE_WINDOW", close)
        self.canvas.bind("<Motion>", self.hover)
        self.canvas.bind("<Leave>", self.leave_window)

        # Run mainloop
        self.root.mainloop()

        # NOTE: The following code == an alternative, but broken, method of running mainloop
        """while self.running:
            self.root.update_idletasks()
            self.root.update()
            sleep(0.01)"""


def Window(
    width=500,
    height=500,
    title="ntk",
    canvas_width="default",
    canvas_height="default",
    closing_command=sys.exit,
    resizable=(True, True),
    override=False,
):
    """Window constructor

    Args:
        width (int, optional): Width. Defaults to 500.
        height (int, optional): Height. Defaults to 500.
        canvas_width (str, optional): Canvas width. Defaults to width.
        canvas_height (str, optional): Canvas height. Defaults to height.
        closing_command (function, optional): Command to execute on close. Defaults to sys.exit.
        resizable (iterable or boolean, optional): Controls whether the window is resizable on the X axis, then Y axis

    Returns:
        _type_: _description_
    """
    # Create window

    if type(resizable) is bool:
        resizable = (resizable, resizable)

    canvas = _window_internal(
        width,
        height,
        title,
        canvas_width,
        canvas_height,
        closing_command,
        resizable,
        override,
    )

    # Start window thread
    canvas.start()

    # Wait for window to be created, as it == in a separate thread and not blocking this thread
    while canvas.root is None:
        sleep(0.1)

    # Return the window
    return canvas


fonts = [
    "Algerian",
    "Blackadder ITC",
    "Bradley Hand ITC",
    "Castellar",
    "Cooper Black",
    "Edwardian Script ITC",
    "Forte",
    "Comic Sans MS",
    "Bauhaus 93",
    "Harlow Solid Italic",
]

colors = [
    "FD3A4A",
    "FD3A4A",
    "A7F432",
    "5DADEC",
    "BFAFB2",
    "FF5470",
    "FFDB00",
    "FF7A00",
    "#45ba52",
    "#2a89d5",
    "#5c21de",
    "#d14d2e",
]

images = [f"examples\\Images\\cloud{i}.png" for i in range(1, 5)]

labels = []


# tween
def animate(widget, target_x, target_y):
    while True:
        sleep(0.01)
        updated_pos = [target_x, target_y]
        if widget.x != target_x:
            updated_pos[0] = widget.abs_x + ((target_x - widget.x) / 10)
        if widget.y != target_y:
            updated_pos[1] = widget.abs_y + ((target_y - widget.y) / 10)
        if updated_pos == [target_x, target_y]:
            break
        widget.place(updated_pos[0], updated_pos[1])


def add_text():
    global entry_text
    global display
    global labels
    global canvas2
    global x
    global y
    global last_random
    print(x, y)

    font_size = random.randint(80, 100)

    font = (
        fonts[random.randint(0, len(fonts) - 1)],
        font_size,
    )

    width, height = fonts_manager.get_min_button_size(
        entry_text.master, font=font, text=entry_text.get()
    )
    if width / height < 820 / 500:
        width = (820 / 500) * height * 1.5
        height = height * 1.5
    else:
        width = width * 1.5
        height = (500 / 820) * width * 1.5
    image = images[random.randint(0, len(images) - 1)]

    iterations = 0
    invalid_location = True
    position = (x, y)
    while invalid_location:
        if last_random:
            position = (random.randint(0, 192) * 10, random.randint(0, 108) * 10)
        label_new = [
            position[0],  # x
            position[1],  # y
            width + position[0],  # x2
            height + position[1],  # y2
        ]
        invalid_location = False
        if label_new[2] > 1920 or label_new[3] > 1080:
            invalid_location = True
        else:
            for label in labels:
                if (
                    label_new[2] >= label[0]
                    and label_new[0] <= label[2]
                    and label_new[1] <= label[3]
                    and label_new[3] >= label[1]
                ):
                    invalid_location = True
                    break
        iterations += 1
        if iterations >= 100:
            font_size -= 5
            font = (
                font[0],
                font_size,
            )
            width, height = fonts_manager.get_min_button_size(
                entry_text.master, font=font, text=entry_text.get()
            )
            if width / height < 820 / 500:
                width = (820 / 500) * height * 1.5
                height = height * 1.5
            else:
                width = width * 1.5
                height = (500 / 820) * width * 1.5
            iterations = 0
    color = colors[random.randint(0, len(colors) - 1)]
    angle = random.randint(-30, 30)
    lbl = Label(
        display,
        text=entry_text.get(),
        height=height,
        width=width,
        image=(image, 50, angle, (255, 255, 255)),
        font=font,
        text_color=color,
        angle=angle,
    )
    lbl.place(random.randint(0, 1920), 1920 + lbl.height)
    Label(
        canvas2,
        text=entry_text.get(),
        height=height / 5,
        width=width / 5,
        # image=(image, 50, (255, 255, 255)),
        font=(font[0], int(font[1] / 5)),
        text_color=color,
    ).place(position[0] / 5, position[1] / 5)
    labels.append(label_new)
    threading.Thread(target=animate, args=(lbl, position[0], position[1])).start()

    x = random.randint(0, 192) * 10
    y = random.randint(0, 108) * 10
    last_random = True


# NOTE: EXAMPLE WINDOW
def __main__():
    for file in os.listdir("examples\\Fonts"):
        print(file)
        fonts_manager.loadfont(file)
    global x
    global y
    global last_random
    last_random = True
    x = random.randint(0, 192) * 10
    y = random.randint(0, 108) * 10

    global display
    display = Window(1920, 1080, resizable=False, override=True)  # .place(1920)
    entry = Window(384, 50 + 216, resizable=False, closing_command=display.close)

    global entry_text
    entry_text = Entry(
        entry,
        text="",
        font=("Helvetica", 20),
        width=300,
        height=50,
        justify="left",
        fill=[0, 100, 0, 50],
        border_width=1,
        text_color="cobaltgreen",
    ).place(0, 0)

    Button(
        entry,
        text="Enter",
        width=50,
        height=50,
        fill=[0, 100, 0, 50],
        command=add_text,
    ).place(300, 0)

    def clicked(event):
        global x
        global y
        global last_random
        if event.y > 50:
            last_random = False
            x, y = event.x * 5, (event.y - 50) * 5

    entry.bind("<Button-1>", clicked)

    Frame(
        display, width=1920, height=1080, image="examples/Images/background2.jpg"
    ).place()

    global canvas2
    canvas2 = Frame(
        entry, width=384, height=216, image="examples/Images/background2.jpg"
    ).place(0, 50)


if __name__ == "__main__":
    __main__()



try:
    from . import image_manager
    from . import fonts_manager
    from . import bounds_manager
    from . import colors_manager
    from . import standard_methods
except ImportError:
    import image_manager
    import fonts_manager
    import bounds_manager
    import colors_manager
    import standard_methods

# ============================================================ INITIALIZATION METHODS ======================================================================


def load_initial(
    _object,
    root,
    width,
    height,
    border_width,
    justify=None,
    state=False,
    mode=None,
    command=None,
    command_off=None,
    slider_height=10,
    maximum=100,
    minimum=0,
    slider_border_width=0,
):
    """Initial loading for nebulatk widgets.

    Args:
        _object (nebulatk.Widget): widget
        root (nebulatk.Window): root window
        width (int): width
        height (int): height
        border_width (int): border width
        justify (str, optional): Justification for text. Defaults to None.
        state (bool, optional): Defaults to False.
        mode (int, optional): Defaults to None.
        command (function, optional): Defaults to None.
        command_off (function, optional): Defaults to None.
        slider_height (int, optional): Defaults to 10.
        maximum (int, optional): Defaults to 100.
        minimum (int, optional): Defaults to 0.
        slider_border_width (int, optional): Defaults to 0.
    """
    # Set root and master, to make parenting widgets to other widgets easier in the future
    _object.root = root
    _object.master = root.master

    if root != root.master:
        root.children.append(_object)

    _object.children = []

    # Set width and height, ensuring they are integers
    _object.width = int(width)
    _object.height = int(height)
    _object.slider_height = int(slider_height)

    _object.angle = 0
    # Set slider minimum and maximum values, ensuring they are integers
    _object.minimum = int(minimum)
    _object.maximum = int(maximum)

    # Set border, ensuring border_width values are integers
    _object.border_width = int(border_width)
    _object.slider_border_width = int(slider_border_width)

    # Initialize text variables
    _object.text = ""
    _object.original_text = ""
    _object.justify = justify

    # Set state and mode
    _object.state = state
    _object.mode = mode

    # Set commands
    _object.command = command
    _object.command_off = command_off

    # Initialize other variables to their defaults
    _object.visible = True
    _object.hovering = False

    _object.x = 0
    _object.y = 0

    _object.original_x = 0
    _object.original_y = 0

    _object.end = 0

    _object.object = None
    _object.slider_object = None
    _object.slider_bg_object = None
    _object.bg_object = None
    _object.bg_object_active = None
    _object.text_object = None
    _object.active_text_object = None
    _object.image_object = None


def load_text(_object, text, font):
    """Initializes text for nebulatk widgets

    Args:
        _object (nebulatk.Widget): widget
        text (str): text
        font (tuple): font tuple
    """
    if text != "":
        # Generate font
        if font is None:
            # Default font
            font = ("Helvetica", int(_object.width / len(text)))

        # If only font name is specified
        if type(font) is str:
            font = (font, int(_object.width / len(text)))

        # Get minimum size of widget
        min_width, min_height = fonts_manager.get_min_button_size(
            _object.master, font, text
        )

        # If the widget size is not specified, set it to the minimum size
        if _object.width == 0:
            _object.width = min_width
        if _object.height == 0:
            _object.height = min_height

        # Check if the font size is the default font size
        # If so, set it to the max font size possible for the widget size
        if font == (font[0], int(_object.width / len(text))):
            font = (
                font[0],
                fonts_manager.get_max_font_size(
                    _object.master, font, _object.width, _object.height, text
                ),
            )

    _object.text = text
    _object.entire_text = text
    _object.font = font


def load_bounds_type(_object, bounds_type, image=None):
    """Initialize bounds_type for nebulatk widgets

    Args:
        _object (nebulatk.Widget): widget
        bounds_type (str): bounds type
        image (_type_, optional): Image. Defaults to None.
    """
    # If the bounds type is the default, determine which default is required
    if bounds_type == "default":
        bounds_type = "box" if image is None else "non-standard"
    _object.bounds_type = bounds_type


def load_bulk_images(_object, **kwargs):
    """Load all images specified

    Args:
        _object (nebulatk.Widget): widget
        kwargs: Image name = image path
    """
    pil_images = {}

    for arg, value in kwargs.items():
        if value is None:
            setattr(_object, arg, None)
            continue
        # Load image
        if type(kwargs[arg]) is str:
            img, pil_img = image_manager.load_image(_object, value, True)
        else:
            _, pil_img = image_manager.load_image(_object, value[0], True)
            pil_img = pil_img.rotate(value[2], expand=True)
            pil_img = pil_img.convert("RGBA")
            data = pil_img.getdata()

            if len(value) < 4:
                new_data = [
                    (
                        *data[i][:3],
                        standard_methods.clamp(data[i][3] - value[1], 0, 255),
                    )
                    for i in range(len(data) - 1)
                ]
            else:
                new_data = [
                    (
                        *value[3],
                        standard_methods.clamp(data[i][3] - value[1], 0, 255),
                    )
                    for i in range(len(data) - 1)
                ]
            pil_img.putdata(new_data)
            """
            im2 = pil_img.copy()
            im2.putalpha(kwargs[arg][1])
            pil_img.paste(im2, pil_img)"""
            img = image_manager.convert_image(_object, pil_img)
        # Load it into the widget
        setattr(_object, arg, img)
        pil_images[arg] = pil_img

    # Generate non-standard bounds if necessary
    if _object.image is not None:
        if _object.bounds_type == "non-standard":
            _object.bounds = bounds_manager.generate_bounds_for_nonstandard_image(
                pil_images["image"]
            )

        # New width and height is the image size
        _object.width, _object.height = pil_images["image"].size
        _object.width += _object.border_width * 2
        _object.height += _object.border_width * 2


def load_all_colors(
    _object,
    fill="default",
    active_fill="default",
    border="default",
    text_color="default",
    active_text_color=None,
    slider_fill=None,
    slider_border=None,
):
    """Load all colors into nebulatk widgets

    Args:
        _object (nebulatk.Widget): widget
        fill (color, optional): Fill color. Defaults to "default".
        active_fill (color, optional): Active fill color. Defaults to "default".
        border (color, optional): Border color. Defaults to "default".
        text_color (color, optional): Text color. Defaults to "default".
        active_text_color (color, optional): Active text color. Defaults to None.
        slider_fill (color, optional): Slider fill color. Defaults to None.
        slider_border (color, optional): Slider border color. Defaults to None.
    """
    # Generate default colors
    if _object.image is None:
        if fill == "default":
            fill = "white"

        if active_fill == "default":
            if fill is None:
                active_fill = None

            elif colors_manager.check_full_white_or_black(fill) == 1:
                active_fill = colors_manager.darken(fill, 40)
            else:
                active_fill = colors_manager.brighten(fill, 40)

        if border == "default":
            border = "black"

    else:
        if fill == "default":
            fill = None
        if active_fill == "default":
            active_fill = None
        if border == "default":
            border = None

    # Set remaining defaults
    if text_color == "default":
        text_color = "black"

    if slider_fill == "default":
        slider_fill = "gray"
    if slider_border == "default":
        slider_border = None

    # Load colors into widget

    _object.fill = fill
    _object.active_fill = active_fill
    _object.slider_fill = slider_fill

    _object.border = border
    _object.slider_border = slider_border

    _object.text_color = text_color
    _object.active_text_color = active_text_color


# Transform all colors to hex for uniformity
def convert_all_colors(_object):
    """Transform all colors to hex for uniformity

    Args:
        _object (nebulatk.Widget): widget
    """
    _object.fill = colors_manager.convert_to_hex(_object.fill)
    _object.active_fill = colors_manager.convert_to_hex(_object.active_fill)
    _object.border = colors_manager.convert_to_hex(_object.border)
    _object.text_color = colors_manager.convert_to_hex(_object.text_color)
    _object.active_text_color = colors_manager.convert_to_hex(_object.active_text_color)
    _object.slider_fill = colors_manager.convert_to_hex(_object.slider_fill)
    _object.slider_border = colors_manager.convert_to_hex(_object.slider_border)



try:
    from . import standard_methods
except ImportError:
    import standard_methods


def generate_bounds_for_nonstandard_image(image, tolerance=0.75):
    """Generates bounds for an image with arbitrary shapes and transparency

    Args:
        image (PilImage): Image to generate bounds for
        tolerance (float, optional): Minimum opacity required to be added to bounds. Defaults to 0.75.

    Returns:
        dict: Resultant bounds
    """
    pixels = image.load()
    width, height = image.size

    bounds = {}
    start_pixel = None

    # Execute function twice if no bounds are found the first pass
    while not bounds and tolerance >= 0:
        # Iterate over the pixels
        for y in range(height):
            for x in range(width):
                cpixel = pixels[x, y]

                # If the pixel is opaque enough to be over the tolerance threshold, and no line start was found, it is the start of a line
                if cpixel[3] >= round(tolerance * 255):
                    if start_pixel is None:
                        start_pixel = x

                # If the pixel doesn't hit the tolerance threshold, the line ends with the previous pixel
                elif start_pixel is not None:
                    # Add line to the bounds
                    if y in bounds:
                        bounds[y].append([start_pixel, x - 1])
                    else:
                        bounds[y] = [[start_pixel, x - 1]]
                    start_pixel = None

            # If no end to the line was found, the line must continue to the end
            if start_pixel is not None:
                bounds[y] = [[start_pixel, x]]
                start_pixel = None

        # In case no bounds were found, reduce tolerance
        tolerance -= tolerance
    return bounds


def remove_bounds(item, mode="box"):
    """Remove the bounds of a widget

    Args:
        item (nebulatk.Widget): widget
        mode (str, optional): Mode to remove bounds with. Defaults to "box".
    """

    x, y = standard_methods.rel_position_to_abs(item, item.x, item.y)

    if mode == "box":
        for i in range(y, y + item.height):
            if i in item.master.bounds:
                for j in range(len(item.master.bounds[i])):
                    if item.master.bounds[i][j][0] == item:
                        item.master.bounds[i].pop(j)
                        break

    elif mode == "non-standard":
        for i in item.bounds.keys():
            for bound in item.bounds[i]:
                if i in item.master.bounds:
                    for j in range(len(item.master.bounds[i])):
                        if (
                            item.master.bounds[i][j][0] == item
                            and item.master.bounds[i][j][1] == bound[0] + x
                        ):
                            item.master.bounds[i].pop(j)
                            break


def update_bounds(item, x, y, mode="box"):
    """Update the bounds of a widget

    Args:
        item (nebulatk.Widget): widget
        x (int): x
        y (int): y
        mode (str, optional): Mode to update bounds with. Defaults to "box".
    """

    # Set x and y in case none is provided
    if x is None:
        x = item.x
    if y is None:
        y = item.y

    x, y = standard_methods.rel_position_to_abs(item, x, y)
    # Remove old bounds
    remove_bounds(item, mode)

    if mode == "box":
        # Add in new bounds
        for i in range(y, y + item.height):
            if i in item.master.bounds:
                if item.master.bounds[i] != []:
                    for j in range(len(item.master.bounds[i])):
                        if item.master.bounds[i][j][0].object < item.object:
                            item.master.bounds[i].insert(j, [item, x, x + item.width])
                            break
                else:
                    item.master.bounds[i] = [[item, x, x + item.width]]
            else:
                item.master.bounds[i] = [[item, x, x + item.width]]

    elif mode == "non-standard":
        # Add in new bounds
        # Iterate over all y values in the new bounds
        for i in item.bounds.keys():

            # Iterate over all lines defined in the new bounds[y]
            for bound in item.bounds[i]:

                # If there is already a list of objects in the master bounds at this y
                if i + y in item.master.bounds:

                    # Check if the list of objects is empty
                    if item.master.bounds[i] != []:

                        # If it isn't empty, iterate over all existing objects in the master bounds until we find an object that the bounds should be behind
                        for j in range(len(item.master.bounds[i + y])):
                            if item.master.bounds[i + y][j][0].object < item.object:
                                item.master.bounds[i + y].insert(
                                    j, [item, bound[0] + x, bound[1] + x]
                                )

                                # Stop looking
                                break

                    # Otherwise, create new list of objects in master bounds
                    else:
                        item.master.bounds[i] = [[item, bound[0] + x, bound[1] + x]]
                else:
                    item.master.bounds[i] = [[item, bound[0] + x, bound[1] + x]]



try:
    from . import image_manager
    from . import fonts_manager
    from . import bounds_manager
    from . import colors_manager
except ImportError:
    import image_manager
    import fonts_manager
    import bounds_manager
    import colors_manager


# Define the names of objects
IMAGE_OBJECTS = [
    "image_object",
    "active_object",
    "hover_object",
    "hover_object_active",
]

BG_OBJECTS = ["bg_object", "bg_object_active"]

TEXT_OBJECTS = [
    "text_object",
    "active_text_object",
]

SLIDER_OBJECTS = ["slider_object", "slider_bg_object"]

ALL_OBJECTS = IMAGE_OBJECTS + BG_OBJECTS + TEXT_OBJECTS + SLIDER_OBJECTS


def check(_object, attribute):
    """Check if the given attribute is defined and is not None.

    Args:
        _object (nebulatk.Widget): widget
        attribute (str): attribute

    Returns:
        bool: Returns True if both checks are satisfied, False otherwise
    """
    return bool(hasattr(_object, attribute) and getattr(_object, attribute) is not None)


# Unfortunately not included in the python math library
def clamp(number, minimum=0, maximum=0):
    """Implement clamp method

    Args:
        number (int): Number to clamp
        minimum (int, optional): Floor. Defaults to 0.
        maximum (int, optional): Ceiling. Defaults to 0.

    Returns:
        int: Clamped number
    """
    number = min(number, maximum)
    number = max(number, minimum)

    return number


def rel_position_to_abs(_object, x, y):
    # Add up positions for relative offsets in parent tree
    obj = _object
    while obj.root != obj.master:
        x += obj.root.x
        y += obj.root.y
        obj = obj.root
    return x, y


# ============================================================ FLOPS ======================================================================
# NOTE: Flops hide all items of a type, and shows the selected item


def image_flop(_object, val):
    """Hides all images, and shows the selected image

    Args:
        _object (nebulatk.Widget): widget
        val (str): Item to show
    """
    visible = "normal" if _object.visible else "hidden"
    if check(_object, val):
        for obj in IMAGE_OBJECTS:
            if hasattr(_object, obj):
                if val == obj:
                    _object.master.change_state(getattr(_object, obj), state=visible)
                else:
                    _object.master.change_state(getattr(_object, obj), state="hidden")


def bg_flop(_object, val):
    """Hides all background objects, and shows the selected background object

    Args:
        _object (nebulatk.Widget): widget
        val (str): Item to show
    """
    visible = "normal" if _object.visible else "hidden"
    for obj in BG_OBJECTS:
        if check(_object, obj):
            if val == obj:
                _object.master.change_state(getattr(_object, obj), state=visible)
            else:
                _object.master.change_state(getattr(_object, obj), state="hidden")


def text_flop(_object, val):
    """Hides all text objects, and shows the selected text object

    Args:
        _object (nebulatk.Widget): widget
        val (str): Item to show
    """
    visible = "normal" if _object.visible else "hidden"
    if hasattr(_object, val) and getattr(_object, val) is not None:
        for obj in TEXT_OBJECTS:
            if hasattr(_object, obj):
                if val == obj:
                    _object.master.change_state(getattr(_object, obj), state=visible)
                else:
                    _object.master.change_state(getattr(_object, obj), state="hidden")


def flop_off(_object):
    """Hides all objects

    Args:
        _object (nebulatk.Widget): widget
    """
    for obj in ALL_OBJECTS:
        if hasattr(_object, obj):
            _object.master.change_state(getattr(_object, obj), state="hidden")


def flop_on(_object):
    """Shows all objects

    Args:
        _object (nebulatk.Widget): widget
    """
    visible = "normal" if _object.visible else "hidden"
    if _object.state:
        if _object.hovering:
            image_flop(_object, "hover_object_active")
        else:
            image_flop(_object, "active_object")
    elif _object.hovering:
        image_flop(_object, "hover_object")
    else:
        image_flop(_object, "image_object")
    if _object.bg_object_active is not None and _object.state:
        bg_flop(_object, "bg_object_active")

    else:
        bg_flop(_object, "bg_object")

    if _object.active_text_object is not None and _object.state:
        text_flop(_object, "active_text_object")
    else:
        text_flop(_object, "text_object")

    _object.master.change_state(_object.slider_bg_object, visible)


# ============================================================ STANDARD WIDGET MANAGMENT METHODS ==========================================


# ------------------------------------------------------------ DELETION BEHAVIOUR -----------------------------------------------------------


def delete(_object):
    """Widget deletion behaviour

    Args:
        _object (nebulatk.Widget): widget
    """
    # Iterate through all items in the widget and delete them if they exist
    for obj in ALL_OBJECTS:
        if check(_object, obj):
            _object.master.delete(getattr(_object, obj))
            setattr(_object, obj, None)


# ------------------------------------------------------------ HOVERING BEHAVIOUR -----------------------------------------------------------


def hovered_standard(_object):
    """Widget hover behaviour for standard type widgets

    Args:
        _object (nebulatk.Widget): widget
    """
    if _object.visible:
        image_flop(_object, "hover_object")
        if _object.bg_object_active is not None:
            bg_flop(_object, "bg_object_active")
        if _object.active_text_object is not None:
            text_flop(_object, "active_text_object")


def hovered_toggle(_object):
    """Widget hover behaviour for toggle type widgets

    Args:
        _object (nebulatk.Widget): widget
    """
    if _object.visible:
        if _object.state:
            image_flop(_object, "hover_object_active")
        else:
            image_flop(_object, "hover_object")


def hover_end(_object):
    """Widget hover end behaviour for all widgets

    Args:
        _object (nebulatk.Widget): widget
    """
    if _object.visible:
        if not _object.state:
            image_flop(_object, "active_object")
        else:
            image_flop(_object, "image_object")
            if _object.active_text_object is not None:
                text_flop(_object, "text_object")
        if _object.bg_object_active is not None:
            bg_flop(_object, "bg_object")


# ------------------------------------------------------------ TOGGLE BEHAVIOUR -------------------------------------------------------------


def toggle_object_standard(_object):
    """Widget toggle behaviour for standard type widgets

    Args:
        _object (nebulatk.Widget): widget
    """
    _object.state = not _object.state
    if _object.hovering:
        image_flop(_object, "hover_object_active")
    else:
        image_flop(_object, "active_object")


def toggle_object_toggle(_object):
    """Widget toggle behaviour for toggle type widgets

    Args:
        _object (nebulatk.Widget): widget
    """
    _object.state = not _object.state
    if _object.state:
        if _object.hovering:
            image_flop(_object, "hover_object_active")
        else:
            image_flop(_object, "active_object")

    elif _object.hovering:
        image_flop(_object, "hover_object")
    else:
        image_flop(_object, "image_object")
    if check(_object, "bg_object_active"):
        if not _object.state:
            bg_flop(_object, "bg_object")

        else:
            bg_flop(_object, "bg_object_active")

    if check(_object, "active_text_object"):
        if not _object.state:
            text_flop(_object, "text_object")

        else:
            text_flop(_object, "active_text_object")


# ------------------------------------------------------------ CLICK BEHAVIOUR --------------------------------------------------------------


def clicked_standard(_object):
    """Widget click behaviour for standard type widgets

    Args:
        _object (nebulatk.Widget): widget
    """
    if _object.visible:
        toggle_object_standard(_object)
        if _object.command is not None:
            _object.command()


def clicked_toggle(_object):
    """Widget click behaviour for toggle type widgets

    Args:
        _object (nebulatk.Widget): widget
    """
    if _object.visible:
        toggle_object_toggle(_object)

        if _object.command is not None and _object.state:
            _object.command()
        if _object.command_off is not None and not _object.state:
            _object.command_off()
        elif _object.command is not None and not _object.state:
            _object.command()


# ------------------------------------------------------------ POSITION BEHAVIOURS ----------------------------------------------------------


def _update_position(_object, item, x, y):
    """Internal update_position method

    Args:
        _object (nebulatk.Widget): widget
        item (str): item
        x (int): x position
        y (int): y position
    """
    old_x, old_y = rel_position_to_abs(_object, _object.x, _object.y)
    if check(_object, item):
        _object.master.move(getattr(_object, item), x - old_x, y - old_y)


def update_positions(_object, x, y, avoid_slider=False):
    """Update positions of all objects

    Args:
        _object (nebulatk.Widget): widget
        x (int): x position
        y (int): y position
        avoid_slider (bool, optional): Request to avoid touching the slider background objects. Defaults to False.
    """
    x, y = rel_position_to_abs(_object, x, y)

    for obj in ALL_OBJECTS:
        if obj == "slider_bg_object" and avoid_slider:
            continue

        _update_position(_object, obj, x, y)


# ------------------------------------------------------------ PLACEMENT BEHAVIOURS ---------------------------------------------------------


def place_bulk(_object, x, y):
    """Bulk place objects

    Args:
        _object (nebulatk.Widget): widget
    """
    x, y = rel_position_to_abs(_object, x, y)
    # Only place background rectangles if there == a fill or border
    # Place slider_bg_object
    state = "normal" if _object.visible else "hidden"
    if _object.slider_fill is not None or (
        _object.slider_border is not None and _object.slider_border_width != 0
    ):
        _object.slider_bg_object = _object.master.create_rectangle(
            x,
            y + _object.height / 2 - _object.slider_height / 2,
            x + _object.maximum + _object.width,
            y + _object.height / 2 - _object.slider_height / 2 + _object.slider_height,
            fill=_object.slider_fill,
            border_width=_object.slider_border_width,
            outline=_object.slider_border,
            state=state,
        )
    # Place bg_object
    if _object.fill is not None or (
        _object.border is not None and _object.border_width != 0
    ):
        _object.bg_object = _object.master.create_rectangle(
            x,
            y,
            x + _object.width,
            y + _object.height,
            fill=_object.fill,
            border_width=_object.border_width,
            outline=_object.border,
            state=state,
        )

    # Place bg_object_active
    if _object.active_fill is not None:
        _object.bg_object_active = _object.master.create_rectangle(
            x,
            y,
            x + _object.width,
            y + _object.height,
            fill=_object.active_fill,
            border_width=_object.border_width,
            outline=_object.border,
            state="hidden",
        )

    # Place images
    for img in ["image", "active_image", "hover_image", "hover_image_active"]:
        if check(_object, img):
            state = "hidden"
            img_object = img.split("_")[0] + "_object"
            if img == "image" and _object.visible:
                state = "normal"
            if img == "hover_image_active":
                img_object += "_active"

            setattr(
                _object,
                img_object,
                _object.master.create_image(
                    x + _object.border_width,
                    y + _object.border_width,
                    getattr(_object, img),
                    state=state,
                ),
            )

    # Place text objects
    if _object.text != "":
        generate_text(_object, x, y)


def generate_text(_object, x, y):
    state = "normal" if _object.visible else "hidden"
    # Set x offset and anchor based on justify
    if _object.justify == "center":
        local_x = x + (_object.width / 2)
        anchor = "center"

    elif _object.justify == "left":
        local_x = x
        anchor = "w"

    elif _object.justify == "right":
        local_x = x + _object.width
        anchor = "e"

    # Set y offset
    local_y = y + (_object.height / 2)

    _object.text_object = _object.master.create_text(
        local_x,
        local_y,
        text=_object.text,
        font=_object.font,
        fill=_object.text_color,
        anchor=anchor,
        state=state,
        angle=_object.angle,
    )
    if _object.active_text_color is not None:
        _object.active_text_object = _object.master.create_text(
            local_x,
            local_y,
            text=_object.text,
            font=_object.font,
            fill=_object.active_text_color,
            anchor=anchor,
            state="hidden",
        )



from tkinter import font as tkfont
import math

ctypes_available = True
try:
    from ctypes import windll, byref, create_unicode_buffer
except Exception as e:
    print(e)
    ctypes_available = False

FR_PRIVATE = 0x10
FR_NOT_ENUM = 0x20


def loadfont(fontpath, private=True, enumerable=False):
    if not ctypes_available:
        return
    """Load a font into resources so that a process can use it

    Args:
        fontpath (str): Path to the font
        private (bool, optional): Whether this font should be private to this process. Defaults to True.
        enumerable (bool, optional): Enumerable. Defaults to False.

    Raises:
        TypeError: Fontpath must be of type str

    Returns:
        bool: Success of loading specified font
    """
    if not isinstance(fontpath, str):
        raise TypeError("Fontpath must be of type str")

    # Generate buffer and load the font resources
    pathbuf = create_unicode_buffer(fontpath)
    AddFontResourceEx = windll.gdi32.AddFontResourceExW
    # Generate flags
    flags = (FR_PRIVATE if private else 0) | (0 if enumerable else FR_NOT_ENUM)

    # Add fonts to resource, and return number of fonts added successfully
    numFontsAdded = AddFontResourceEx(byref(pathbuf), flags, 0)

    # Return success based on number of fonts added
    return bool(numFontsAdded)


def get_max_font_size(root, font, width, height, text):
    """Find the maximum font size for a given font and dimensions.

    Args:
        root (nebulatk.Window): The base window created by nebulatk.Window()
        font (tuple): A tuple containing the font name, size, and optionally a style
        width (int): The width of the bounding rectangle
        height (int): The height of the bounding rectangle
        text (str): The text that needs to fit in the bounding rectangle

    Returns:
        int: The maximum font size
    """

    # Set size to 90% of total widget size
    width *= 0.9
    height *= 0.9

    # The code will generate the maximum font size that can fit in the width, and a separate font size for the height

    # This size will be the font size for the width
    size = 1
    prev_size = 0

    # This size will be the font size for the width
    size2 = 1
    prev_size2 = 0

    # Generate full length font tuple
    if len(font) < 3:
        font = (font[0], font[1], tkfont.NORMAL)

    # Generate starting width of text given the font size
    curr_width = tkfont.Font(
        root.root, font=(font[0], size, font[2]), text=text
    ).measure(text)

    # Generate starting height of text given the font size2
    curr_height = tkfont.Font(
        root.root, font=(font[0], size, font[2]), text=text
    ).metrics("linespace")

    # Gradually increase size until the width of the text exceeds the bounds
    while curr_width < width:
        prev_size = size
        size += 1
        curr_width = tkfont.Font(
            root.root, font=(font[0], size, font[2]), text=text
        ).measure(text)

    # Gradually increase size2 until the height of the text exceeds the bounds, or until size2 reaches size1
    while curr_height < height and size2 <= prev_size:
        prev_size2 = size2
        size2 += 1
        curr_height = tkfont.Font(
            root.root, font=(font[0], size2, font[2]), text=text
        ).metrics("linespace")

    # Final font
    # font = tkfont.Font(root.root, font=(font[0], prev_size2, font[2]), text=text)

    # Return last valid size
    return prev_size2


def get_min_button_size(root, font, text):
    """Find the minimum size of a button with a specified font and text. Returns width, height.

    Args:
        root (nebulatk.Window): The base window created by nebulatk.Window()
        font (tuple): A tuple containing the font name, size, and optionally a style
        text (str): The text that needs to fit in the bounding rectangle

    Returns:
        width (int): The minimum width of the button
        height (int): The minimum height of the button
    """

    # Generate full length font tuple
    if len(font) < 3:
        font = (font[0], font[1], tkfont.NORMAL)

    # Minimum width is 110% of the width of the given font and text
    width = int(
        math.ceil(tkfont.Font(root.root, font=font, text=text).measure(text) / 0.9)
    )

    # Minimum height is 110% of the height of the given font and text
    height = int(
        math.ceil(
            tkfont.Font(root.root, font=font, text=text).metrics("linespace") / 0.9
        )
    )

    return width, height


def get_max_length(root, text, font, width, end=0):
    """Generate the maximum slice of text that can fit in a fixed width

    Args:
        root (nebulatk.Window): root window
        text (str): text
        font (tuple): font
        width (int): width
        end (int, optional): Slice end position. Defaults to 0.

    Returns:
        _type_: _description_
    """

    # Generate full length font tuple
    if len(font) < 3:
        font = (font[0], font[1], tkfont.NORMAL)

    # Initialize length of slice to be the maximum length of the text
    length = len(text)

    # Get the current width of the slice
    curr_width = tkfont.Font(
        root.root, font=font, text=text[end - length : end]
    ).measure(text[end - length : end])

    # Decrease length of slice until it fits
    while curr_width > width and length > 0:
        length -= 1
        curr_width = tkfont.Font(
            root.root, font=font, text=text[end - length : end]
        ).measure(text[end - length : end])

    # Return slice length
    return length


# Symbol sets

ALPHA = [
    "a",
    "b",
    "c",
    "d",
    "e",
    "f",
    "g",
    "h",
    "i",
    "j",
    "k",
    "l",
    "m",
    "n",
    "o",
    "p",
    "q",
    "r",
    "s",
    "t",
    "u",
    "v",
    "w",
    "x",
    "y",
    "z",
]

i_e = len(ALPHA)
ALPHA.extend(str.upper(ALPHA[i]) for i in range(i_e))
NUMERIC = list("12345567890")
SYMBOL = list("`~!@#$%^&*()_+-={}[]|\:\";',.></? ")
ALPHANUMERIC = ALPHA + NUMERIC
ALPHANUMERIC_PLUS = ALPHANUMERIC + SYMBOL



from PIL import Image as pil
from PIL import ImageTk as piltk
from PIL import ImageDraw as pildraw


def convert_image(_object, image):
    return piltk.PhotoImage(image, master=_object.master.root)


def load_image(_object, image, return_both=False):
    """Load an image with PIL

    Args:
        _object (nebulatk.Widget): widget
        image (str): path to the image
        return_both (bool, optional): Request for both TkImage and PilImage. Defaults to False.

    Returns:
        TkImage: Tkinter-compatible image
    OR:
        TkImage: Tkinter-compatible image
        PilImage: Pil image
    """
    image_converted = None
    if image is not None:
        # Open image
        image = pil.open(image)

        # Resize image if size isn't specified
        if _object.width != 0 and _object.height != 0:
            image = image.resize(
                (
                    _object.width - (_object.border_width * 2),
                    _object.height - (_object.border_width * 2),
                ),
                pil.NEAREST,
            )

        # Convert image for tkinter
        image_converted = convert_image(_object, image)

    # Return both PhotoImage and PilImage objects if requested
    return (image_converted, image) if return_both else image_converted


def load_image_generic(window, image, return_both=False):
    """Alternative to load_image without resizing. Does not require a parent widget.

    Args:
        image (str): path to the image
        return_both (bool, optional): Request for both TkImage and PilImage. Defaults to False.

    Returns:
        TkImage: Tkinter-compatible image
    OR:
        TkImage: Tkinter-compatible image
        PilImage: Pil image
    """
    image_converted = None
    if image is not None:
        # Open image
        image = pil.open(image)

        # Convert image for tkinter
        image_converted = convert_image(image, window)

    # Return both PhotoImage and PilImage objects if requested
    return (image_converted, image) if return_both else image_converted


def create_image(fill, width, height, border, border_width, master):
    """Generate a new TkImage image

    Args:
        fill (color): fill color
        width (int): width
        height (int): height
        border (color): border color
        border_width (int): border_width

    Returns:
        TkImage: Tkinter-compatible image
    """
    # Create base image
    image = pil.new("RGBA", (width, height), (0, 0, 0, 255))

    # Generate corners of rectangle (0-indexed)
    shape = [(0, 0), (width - 1, height - 1)]

    # Draw rectangle on image
    image1 = pildraw.Draw(image)
    image1.rectangle(shape, fill=fill, outline=border, width=border_width)

    # Convert image
    image = convert_image(master, image)
    return image
